var topic = "<!-- saved from url=(0024)http://docs.autodesk.com -->\n\
<html>\n\
   <head>\n\
<link href=\"../../style/prettify.css\" type=\"text/css\" rel=\"stylesheet\">\n\
<script type=\"text/javascript\" src=\"../../scripts/prettify.js\"></script><script src=\"../../scripts/lib/jquery-1.11.1.min.js\" type=\"text/javascript\"></script><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"><meta http-equiv=\"Content-Style-Type\" content=\"text/css\"><meta name=\"generator\" content=\"pandoc\"><meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"><script type=\"text/javascript\" src=\"../../scripts/utils/adsk.redirect.js\"></script>\n\
      <title>OpenMP</title>\n\
   <meta name=\"topic-subtype\" content=\"C++\"></head>\n\
   <body height=\"100%\"><div class=\"body_content\" id=\"body-content\"><style type=\"text/css\">code{white-space: pre;}</style><script>$(document).ready(function() { yepnope.injectJs(\"./scripts/multireflink.js\"); });</script><script>$(document).ready(function () { prettyPrint(); } );</script><script></script><script></script><div id=\"reflinkdiv\"></div>\n\
      <div>\n\
         <div class=\"head\">\n\
            <h1>OpenMP</h1>\n\
         </div>\n\
\n\
<div class=\'section\'><a id=\"openmp\"></a></div>\n\
<div class=\'section\'><a id=\"openmp-\"></a><h2 id=\"openmp-\">OpenMP の概要</h2></div>\n\
<p>OpenMP は、主にデータの並列処理用に設計された高レベルのスレッド抽象であり、<strong>for</strong> ループに適用されます。これは、VS2005 とすべてのプラットフォームの Intel コンパイラでサポートされます。OSX 版と Linux 版の Maya で必要な gcc のバージョンは、OpenMP をサポートしていません。</p>\n\
<p>OpenMP はコンパイラ プラグマを介して実装されます。ネイティブ スレッドと比較すると、非常にシンプルなインタフェースです。スレッドの作成と管理のすべてとデータ パーティション作成の大部分は、ユーザに対して表示されません。OpenMP は、初めて並列領域が発生したときに作成されるスレッド プールを使用します(このため、OpenMP コードのプロファイル時にアプリケーションで最初のループ トラバースを無視することが重要です)。</p>\n\
<p>以下に、スレッド化ループの簡単な例を示します。</p>\n\
<div class=\"codeBlock\"><pre class=\"prettyprint\">#pragma omp parallel for\n\
for(int i=0; i&lt;imax; i++) {\n\
    doThreadsafeWork(i);\n\
}\n\
</pre></div><p>このコードは、コンパイラによって変換され、複数スレッド(既定では論理プロセッサごとに 1 スレッド)によって並列で実行される関数になります。最初のプロセッサはループの最初の(imax/numThreads)要素を取り、2 番目のプロセッサは次の同等サイズのチャンク、というふうに続きます。これにより、優れた<strong>キャッシュ アフィニティ</strong>が実現し、プロセッサ通常は隣接するデータ要素を処理するため、キャッシュ ミスを最小限に抑ることができます。</p>\n\
<p>以下に、より複雑な例を示します。</p>\n\
<div class=\"codeBlock\"><pre class=\"prettyprint\">#pragma omp parallel for schedule(guided) if(imax&gt;1000)\n\
for(int i=0; i&lt;imax; i++) {\n\
    doThreadsafeWork(i);\n\
    #pragma omp critical\n\
    doNonThreadsafeWork(i);\n\
}\n\
</pre></div><p>このコードもループをチャンクに分割しますが、guided スケジュール設定オプションによってサイズ(imax/numThreads)よりも小さいチャンクが使用され、既存のチャンクが完了すると新しいチャンクが各スレッドに送られます。これにより、ループの繰り返し間で作業負荷が異なる場合に、実行時に追加チャンクが早期に完了するスレッドに割り当てられるため、<strong>負荷のバランス</strong>が改善されます。<span class=\'code\'>critical</span> プラグマは、後続ラインにロックを配置します。</p>\n\
<p>メイン プラグマの <span class=\'code\'>if</span> 条件によって、トリップ数(ループの繰り返し回数)が指定した値を超えた場合にのみ、ループは並列で実行されます。並列領域のコールにはオーバーヘッドがあるため、トリップ数が小さすぎるときに評価を並列にしても意味はありません。有用な経験則として、並列領域の開始と終了に 1 万クロック サイクルのオーバーヘッドを想定し、カットオフ トリップ数がこの処理を上回るようにすることをお勧めします。これは特に Maya のようなアプリケーションで重要です。Maya では、単一の非常に複雑なオブジェクトでも多数の非常に単純なオブジェクトでも同じアルゴリズムが実行される場合があります。後者の場合、数千もの非常に短いスレッド化評価のため余計な起動時オーバーヘッドが生じてスレッド化の利点がすべて帳消しになり、カットオフのないスレッド化コードで大幅に処理速度が低下する可能性があります。スレッド化が条件によって中断されない場合でも、ループ内部のコードはコンパイラによって個別の関数に抽出されるため、関数コールの影響を受けます。</p>\n\
<div class=\'section\'><a id=\"openmp-\"></a><h2 id=\"openmp-\">OpenMP の利点と欠点</h2></div>\n\
<p>OpenMP の利点は、クロスプラットフォームのサポート、実装と削除が容易であることです。プラグマを無効にするだけでコードは順次形式に戻るため、動作が同じであることをすばやく確認することができます。上記の例のような非常に単純な OpenMP プラグマを使用すると、非常に大量なコードのスレッド化が可能です。異なるスレッド化ライブラリを使用して最終実装を行う場合でも、スレッド化の利点の迅速なプロトタイピングや評価に非常に便利です。</p>\n\
<p>欠点は、アルゴリズムの適用に制約があり、実装間に互換性がないことです。VS2005 と Intel OpenMP ライブラリは併用できますが、相互の実装は認識されません。したがって、Intel コンパイラでコンパイルした関数をコールする VC2005 でコンパイルしたスレッド化ループでは、Intel コンパイラで定義した OpenMP ロックはすべて無視されます。</p>\n\
<p>OpenMP では、ネストされたスレッド化にも問題があります。この場合、より高レベルのスレッドが生成され、それ自体がより低レベルにスレッド化されるコードを呼び出します。これにより、システムのコアよりも多くのスレッドがアクティブになります。これは、<strong>オーバーサブスクリプション</strong>と呼ばれ、パフォーマンスの低下を引き起こします。アプリケーションにスレッド化が追加されるにつれて、オーバーサブスクリプションの問題は大きくなります。これは、並列でコールされている関数自体がスレッド化されていることに開発者が気付いていない場合があるためです。</p>\n\
<p><strong>オーバーサブスクリプション</strong>は、評価マネージャが並列評価を行う際には特に問題となります。</p>\n\
<div class=\'section\'><a id=\"openmp--maya-\"></a><h2 id=\"openmp--maya-\">OpenMP と Maya のスピンウェイト時間</h2></div>\n\
<p>Maya の一部のアルゴリズム(流体やダイナミクス ソルバを含む)は、OpenMP を使用してスレッド化されています。ほとんどのデータ並列ループは Intel® Threading Building Blocks (TBB)を使用するため、評価マネージャを使用する際の問題を回避しています。「<a href=\'#!/url=./developer/Multithreading-plug-ins/Intel-Threading-Building-Blocks.html\' title=\'\'>Intel Threading Building Blocks</a>」を参照してください。</p>\n\
<p>通常、スピンウェイト時間はゼロに設定し、変更すべきではありません。プラグインで Intel OpenMP を使用し、スピンウェイト時間にゼロ以外の値を設定する場合は、Intel コンパイラ バージョンのヘッダ ファイル <span class=\'code\'>omp.h</span> をインクルードすることにより <span class=\'code\'>kmp_set_blocktime()</span> 関数を使用できます。</p>\n\
<p>ただしこれを行う場合、評価マネージャの並列評価の実行中にブロック時間を設定すると、プラグイン間で潜在的な競合状態が発生することにご注意ください。このような潜在的な競合状態を回避するには、評価の実行中ではなく、プラグインの初期化中にブロック時間を設定するようにします。また、プラグインを終了する前にスピンウェイト時間をゼロに戻し、コントロールを Maya に返すことも忘れないでください: <span class=\'code\'>kmp_set_blocktime(0);</span></p>\n\
<p>Maya は、すべてのプラットフォームの OpenMP に Intel コンパイラを使用します。</p>\n\
      <div class=\"footer-block\"><a href=\"../html/ac.cmtdialog.htm\" class=\"comments-anchor\" target=\"_blank\"><span class=\"comments-link\">Please send us your comment about this page</span></a></div></div>\n\
   </div></body>\n\
</html>\n\
";